@@ -4,7 +4,7 @@
  * LXQt - a lightweight, Qt based, desktop toolset
  * https://lxqt.org
  *
- * Copyright: 2015 LXQt team
+ * Copyright: 2015-2018 LXQt team
  * Authors:
  *   Palo Kisa <palo.kisa@gmail.com>
  *
@@ -36,12 +36,16 @@
 #include <QSocketNotifier>
 #include <QDebug>
 #include <QThread>
+#include <QProcessEnvironment>
+#include <QTimer>
 #include <pty.h>
 #include <unistd.h>
 #include <memory>
 #include <csignal>
 #include <sys/wait.h>
 #include <fcntl.h>
+#include <iostream>
+#include <thread>
 
 namespace
 {
@@ -80,11 +84,55 @@ namespace
             << QObject::tr("%1 version %2\n").arg(app_master).arg(app_version);
     }
 
+    //Note: array must be sorted to allow usage of binary search
+    static constexpr char const * const ALLOWED_VARS[] = {
+        "DISPLAY"
+            , "LANG", "LANGUAGE", "LC_ADDRESS", "LC_ALL", "LC_COLLATE", "LC_CTYPE", "LC_IDENTIFICATION", "LC_MEASUREMENT"
+            , "LC_MESSAGES", "LC_MONETARY", "LC_NAME", "LC_NUMERIC", "LC_PAPER", "LC_TELEPHONE", "LC_TIME"
+            , "PATH", "QT_PLATFORM_PLUGIN", "QT_QPA_PLATFORMTHEME", "WAYLAND_DISPLAY", "XAUTHORITY"
+    };
+    static constexpr char const * const * const ALLOWED_END = ALLOWED_VARS + sizeof (ALLOWED_VARS) / sizeof (ALLOWED_VARS[0]);
+    struct assert_helper
+    {
+        assert_helper()
+        {
+            Q_ASSERT(std::is_sorted(ALLOWED_VARS, ALLOWED_END
+                        , [] (char const * const a, char const * const b) { return strcmp(a, b) < 0; }));
+        }
+    };
+    assert_helper h;
+
     inline void env_workarounds()
     {
-        //cleanup environment
-        //pcmanfm-qt will not start if the DBUS_SESSION_BUS_ADDRESS is preserved
-        unsetenv("DBUS_SESSION_BUS_ADDRESS");
+        std::cerr << LXQTSUDO << ": Stripping child environment except for: ";
+        std::copy(ALLOWED_VARS, ALLOWED_END - 1, std::ostream_iterator<const char *>{std::cerr, ", "});
+        std::cerr << *(ALLOWED_END - 1) << '\n'; // printing the last separately to avoid trailing comma
+        // cleanup environment, because e.g.:
+        // - pcmanfm-qt will not start if the DBUS_SESSION_BUS_ADDRESS is preserved
+        // - Qt apps may change user's config files permissions if the XDG_* are preserved
+        for (auto const & key : QProcessEnvironment::systemEnvironment().keys())
+        {
+            auto const & i = std::lower_bound(ALLOWED_VARS, ALLOWED_END, key, [] (char const * const a, QString const & b) {
+                return b > a;
+            });
+            if (i == ALLOWED_END || key != *i)
+            {
+                unsetenv(key.toLatin1().data());
+            }
+        }
+    }
+
+    inline QString quoteShellArg(const QString& arg, bool userFriendly)
+    {
+        QString rv = arg;
+
+        //^ check if thre are any bash special file characters
+        if (!userFriendly || arg.contains(QRegExp("(\\s|[][!\"#$&'()*,;<=>?\\^`{}|~])"))) {
+            rv.replace(QStringLiteral("'"), QStringLiteral("'\\''"));
+            rv.prepend ('\'').append('\'');
+        }
+
+        return rv;
     }
 }
 
@@ -141,61 +189,99 @@ int Sudo::main()
         //we were invoked through unknown link (or renamed binary)
         usage(tr("%1: no backend chosen!").arg(app_master));
         return 1;
-    } else if (BACK_SU == mBackend && 1 < mArgs.size())
-    {
-        QString cmd = mArgs.replaceInStrings(QRegExp(QStringLiteral("^(.*)$")), "'\\1'").join(QStringLiteral(" "));
-        QTextStream(stderr) << tr("%1: warning - got multiple arguments for %2 backend, squashing into one: %3")
-            .arg(app_master).arg(su_prog).arg(cmd);
-        mArgs.erase(++mArgs.begin(), mArgs.end());
-        mArgs[0] = std::move(cmd);
     }
 
-    mDlg.reset(new PasswordDialog{mArgs});
-    mDlg->setModal(true);
-    lxqtApp->setActiveWindow(mDlg.data());
-
     mChildPid = forkpty(&mPwdFd, nullptr, nullptr, nullptr);
     if (0 == mChildPid)
-        child(); //never returns
-    else if (-1 == mChildPid)
+    {
+        child(); // never returns
+        return 1; // but for sure
+    }
+
+    mDlg.reset(new PasswordDialog{squashedArgs(/*userFriendly = */ true), backendName()});
+    mDlg->setModal(true);
+    lxqtApp->setActiveWindow(mDlg.data());
+
+    if (-1 == mChildPid)
         QMessageBox(QMessageBox::Critical, mDlg->windowTitle()
-                , tr("Failed to fork: %1").arg(strerror(errno)), QMessageBox::Ok).exec();
+                , tr("Syscall error, failed to fork: %1").arg(strerror(errno)), QMessageBox::Ok).exec();
     else
         return parent();
 
     return 1;
 }
 
+QString Sudo::squashedArgs(bool userFriendly) const
+{
+    QString rv;
+
+    rv = quoteShellArg (mArgs[0], userFriendly);
+    for (auto argP = ++mArgs.begin(); argP != mArgs.end(); ++argP) {
+        rv.append (' ').append(quoteShellArg (*argP, userFriendly));
+    }
+
+    return rv;
+}
+
+QString Sudo::backendName (backend_t backEnd)
+{
+    QString rv;
+    // Remove leading paths in case variables are set with full path
+    switch (backEnd) {
+        case BACK_SU   : rv = su_prog;   break;
+        case BACK_SUDO : rv = sudo_prog; break;
+        //: shouldn't be actually used but keep as short as possible in translations just in case.
+        case BACK_NONE : rv = tr("unset");
+    }
+
+    return rv;
+}
 
 void Sudo::child()
 {
-    int params_cnt = 2 //1. su/sudo & last nullptr
-        + 1 //-c for su | -E for sudo
-        + mArgs.size();
+    int params_cnt = 3 //1. su/sudo & "shell command" & last nullptr
+        + (BACK_SU == mBackend ? 1 : 3); //-c for su | -E /bin/sh -c for sudo
     std::unique_ptr<char const *[]> params{new char const *[params_cnt]};
     const char ** param_arg = params.get() + 1;
 
-    std::string program;
-    if (BACK_SU == mBackend)
+    std::string program = backendName().toLocal8Bit().data();
+    if (BACK_SUDO == mBackend)
     {
-        program = su_prog.toStdString();
-        *(param_arg++) = "-c"; //run command
-    } else
-    {
-        program = sudo_prog.toStdString();
         *(param_arg++) = "-E"; //preserve environment
+        *(param_arg++) = "/bin/sh";
     }
+    *(param_arg++) = "-c"; //run command
 
     params[0] = program.c_str();
 
-    std::vector<std::string> arguments;
-    for (const auto & a : mArgs)
-        arguments.push_back(a.toStdString());
-    for (const auto & a : arguments)
-        *(param_arg++) = a.c_str();
+    // Note: we force the su/sudo to communicate with us in the simplest
+    // locale and then set the locale back for the command
+    char const * const env_lc_all = getenv("LC_ALL");
+    std::string command;
+    if (env_lc_all == nullptr)
+    {
+        command = "unset LC_ALL; ";
+    } else
+    {
+        // Note: we need to check if someone is not trying to inject commands
+        // for privileged execution via the LC_ALL
+        if (nullptr != strchr(env_lc_all, '\''))
+        {
+            QTextStream{stderr, QIODevice::WriteOnly} << tr("%1: Detected attempt to inject privileged command via LC_ALL env(%2). Exiting!\n").arg(app_master).arg(env_lc_all);
+            exit(1);
+        }
+        command = "LC_ALL='";
+        command += env_lc_all;
+        command += "' ";
+    }
+    command += "exec ";
+    command += squashedArgs().toLocal8Bit().data();
+    *(param_arg++) = command.c_str();
 
     *param_arg = nullptr;
 
+    setenv("LC_ALL", "C", 1);
+
     env_workarounds();
 
     setsid(); //session leader
@@ -208,7 +294,6 @@ void Sudo::child()
 
 void Sudo::stopChild()
 {
-    close(mPwdFd);
     kill(mChildPid, SIGINT);
     int res, status;
     for (int cnt = 10; 0 == (res = waitpid(mChildPid, &status, WNOHANG)) && 0 < cnt; --cnt)
@@ -217,12 +302,7 @@ void Sudo::stopChild()
     if (0 == res)
     {
         kill(mChildPid, SIGKILL);
-        mRet = 1;
-    } else
-    {
-        mRet = WIFEXITED(status) ? WEXITSTATUS(status) : 1;
     }
-    mChildPid = -1;
 }
 
 int Sudo::parent()
@@ -231,7 +311,7 @@ int Sudo::parent()
     if (0 != fcntl(mPwdFd, F_SETFL, O_NONBLOCK))
     {
         QMessageBox(QMessageBox::Critical, mDlg->windowTitle()
-                , tr("Failed to set non-block: %1").arg(strerror(errno)), QMessageBox::Ok).exec();
+                , tr("Syscall error, failed to bring pty to non-block mode: %1").arg(strerror(errno)), QMessageBox::Ok).exec();
         return 1;
     }
 
@@ -239,7 +319,7 @@ int Sudo::parent()
     if (nullptr == pwd_f)
     {
         QMessageBox(QMessageBox::Critical, mDlg->windowTitle()
-                , tr("Failed to fdopen: %1").arg(strerror(errno)), QMessageBox::Ok).exec();
+                , tr("Syscall error, failed to fdopen pty: %1").arg(strerror(errno)), QMessageBox::Ok).exec();
         return 1;
     }
 
@@ -254,27 +334,24 @@ int Sudo::parent()
             } else
             {
                 stopChild();
-                lxqtApp->quit();
             }
         });
 
     QString last_line;
     QScopedPointer<QSocketNotifier> pwd_watcher{new QSocketNotifier{mPwdFd, QSocketNotifier::Read}};
-    QObject::connect(pwd_watcher.data(), &QSocketNotifier::activated, [&]
+    auto reader = [&]
         {
             QString line = child_str.readAll();
             if (line.isEmpty())
             {
-                pwd_watcher.reset(nullptr);
-                QString const & prog = BACK_SU == mBackend ? su_prog : sudo_prog;
+                pwd_watcher.reset(nullptr); //stop the notifications events
+
+                QString const & prog = backendName();
                 if (last_line.startsWith(QStringLiteral("%1:").arg(prog)))
                 {
-                    pwd_watcher.reset(nullptr); //stop the notifications events
-                    stopChild();
                     QMessageBox(QMessageBox::Critical, mDlg->windowTitle()
                             , tr("Child '%1' process failed!\n%2").arg(prog).arg(last_line), QMessageBox::Ok).exec();
                 }
-                lxqtApp->quit();
             } else
             {
                 if (line.endsWith(pwd_prompt_end))
@@ -296,16 +373,30 @@ int Sudo::parent()
                 last_line = lines.isEmpty() ? QString() : lines.back();
             }
 
+        };
+
+    QObject::connect(pwd_watcher.data(), &QSocketNotifier::activated, reader);
+
+    std::unique_ptr<std::thread> child_waiter;
+    QTimer::singleShot(0, [&child_waiter, this] {
+            child_waiter.reset(new std::thread{[this] {
+                    int res, status;
+                    res = waitpid(mChildPid, &status, 0);
+                    mRet = (mChildPid == res && WIFEXITED(status)) ? WEXITSTATUS(status) : 1;
+                    lxqtApp->quit();
+                }
+            });
         });
 
     lxqtApp->exec();
 
-    if (0 < mChildPid)
-    {
-        int res, status;
-        res = waitpid(mChildPid, &status, 0);
-        mRet = (mChildPid == res && WIFEXITED(status)) ? WEXITSTATUS(status) : 1;
-    }
+    child_waiter->join();
+
+    // try to read the last line(s)
+    reader();
+
+    fclose(pwd_f);
+    close(mPwdFd);
 
     return mRet;
 }
